"""A set of types as required by the DB API 2.0.

All these types implement str() in a way that makes them correct and
safe for inclusion in commands passed to the backend.

This module uses datetime from the stdlib, which means we need Python 2.3+
"""
# System imports
import datetime
import string
import re

# Sibling imports
from errors import *

# Single quote isn't "printable" for us.. we need to muck with it
_printable = string.printable.replace("'","").replace("\\","")

class NULL:
	def __str__(self):
		return "NULL"

class STRING(str):
	def __str__(self):
		s = str.__str__(self)
		s = s.replace("\\","\\\\134")
		return "'" + s.replace("'","''")  + "'"

class BINARY(str):
	def __str__(self):
		return "'" + "".join( [
		(char in _printable and 
		char or
		"\\\\%03o" % ord(char)) 
		for char in str.__str__(self)]) + "'"

class NUMBER: 
	def __init__(self,v):
		if type(v) in (int,float):
			self.v = v		
			return

		raise DataError, ("Cannot convert '%s' to number (int,float)" % v)
	
	def __str__(self):
		return str(self.v)

	def __repr__(self):
		return repr(self.v)

	def __int__(self):
		return int(self.v)

	def __long__(self):
		return long(self.v)

	def __float__(self):
		return float(self.v)

	def __complex__(self):
		return complex(self.v)
	

class ROWID(NUMBER): pass

class DATETIME(datetime.datetime):
	def __init__(self,*args):
		datetime.datetime.__init__(self,*args)

		self._dt = 1
		self._tm = 1

	def __str__(self):
		tot = []
		if self._dt:
			tot.append(self._dodate())
		if self._tm:
			tot.append(self._dotime())
		return "'" + " ".join(tot) + "'"

	def _dodate(self):
		return "%04d-%02d-%02d" % (self.year,self.month,self.day)

	def _dotime(self):
		return "%02d:%02d:%02d.%06d" % (self.hour, self.minute, self.second, self.microsecond)

class BOOL:
	def __init__(self, b):
		self.__b = b
	def __str__(self):
		if self.__b:
			return "'T'"
		else:
			return "'F'"



def Date(y,m,d): 
	dt = DATETIME(y,m,d)
	dt._tm = 0
	return dt

def Time(h,m,s,ms=0): 
	dt = DATETIME(1,1,1,h,m,s,ms)
	dt._dt = 0
	return dt

def Timestamp(y,mo,d,h,m,s,ms=0): return DATETIME(y,mo,d,h,m,s,ms)

def DateFromTicks(t):
	dt = DATETIME.fromtimestamp(t)
	dt._tm = 0
	return dt

def TimeFromTicks(t):
	dt = DATETIME.fromtimestamp(t)
	dt._dt = 0
	return dt

def TimestampFromTicks(t): return DATETIME.fromtimestamp(t)

def Binary(s): return BINARY(s)

def convertUnicode(s):
	return s.decode('utf-8')

_nchar_exp = re.compile('\\\\([0-9]{3})', re.MULTILINE)
def convertBinary(s):
	return re.sub(_nchar_exp, lambda mo: chr(int(mo.group(1),8)),s)

def Bool(b): return BOOL(b)
